@@ -65,7 +65,7 @@
 
 #define CHUNKSIZE 4096
 
-#define oggz_off_t long
+#define oggz_off_t off_t
 
 /*
  * The typical usage is:
