@@ -32,6 +32,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <getopt.h>
 #include <errno.h>
 
@@ -45,6 +46,13 @@
 
 typedef ogg_int64_t timestamp_t;
 
+typedef struct _OVData {
+  OGGZ * writer;
+  OggzTable * missing_eos;
+  int theora_count;
+  int audio_count;
+} OVData;
+
 typedef struct {
   int error;
   char * description;
@@ -128,14 +136,95 @@ gp_to_time (OGGZ * oggz, long serialno, ogg_int64_t granulepos)
   return (timestamp_t)((double)(granule * gr_d) / (double)gr_n);
 }
 
+static int
+read_page (OGGZ * oggz, const ogg_page * og, long serialno, void * user_data)
+{
+  OVData * ovdata = (OVData *)user_data;
+  ogg_int64_t gpos = ogg_page_granulepos((ogg_page *)og);
+  const char * content_type;
+  int ret = 0;
+
+  if (ogg_page_bos ((ogg_page *)og)) {
+    content_type = ot_page_identify (og, NULL);
+    if (!strcmp (content_type, "Theora")) {
+      ovdata->theora_count++;
+      if (ovdata->audio_count > 0) {
+        log_error ();
+        fprintf (stderr, "serialno %010ld: Theora video bos page after audio bos page\n", serialno);
+      }
+    } else if (!strcmp (content_type, "Vorbis") || !strcmp (content_type, "Speex")) {
+      ovdata->audio_count++;
+    }
+  }
+
+  if(gpos != -1 && ogg_page_packets((ogg_page *)og) == 0) {
+    ret = log_error ();
+    fprintf (stderr, "serialno %010ld: granulepos %lld on page with no completed packets, must be -1\n", serialno, gpos);
+  }
+
+  return ret;
+}
+
+static void
+ovdata_init (OVData * ovdata)
+{
+  current_timestamp = 0;
+
+  if ((ovdata->writer = oggz_new (OGGZ_WRITE|OGGZ_AUTO)) == NULL) {
+    fprintf (stderr, "oggz-validate: unable to create new writer\n");
+    exit (1);
+  }
+
+  ovdata->missing_eos = oggz_table_new ();
+  ovdata->theora_count = 0;
+  ovdata->audio_count = 0;
+}
+
+static void
+ovdata_clear (OVData * ovdata)
+{
+  long serialno;
+  int i, nr_missing_eos = 0;
+
+  oggz_close (ovdata->writer);
+
+  if (nr_errors <= MAX_ERRORS) {
+    nr_missing_eos = oggz_table_size (ovdata->missing_eos);
+    for (i = 0; i < nr_missing_eos; i++) {
+      log_error ();
+      oggz_table_nth (ovdata->missing_eos, i, &serialno);
+      fprintf (stderr, "serialno %010ld: missing *** eos\n", serialno);
+    }
+  }
+
+  oggz_table_delete (ovdata->missing_eos);
+}
+
 static int
 read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
 {
-  OGGZ * writer = (OGGZ *)user_data;
+  OVData * ovdata = (OVData *)user_data;
   timestamp_t timestamp;
   int flush;
   int ret = 0, feed_err = 0, i;
 
+  if (op->b_o_s) {
+    oggz_table_insert (ovdata->missing_eos, serialno, (void *)0x1);
+  }
+
+  if (oggz_table_lookup (ovdata->missing_eos, serialno) == NULL) {
+    ret = log_error ();
+    fprintf (stderr, "serialno %010ld: missing *** bos\n", serialno);
+  }
+
+  if (op->e_o_s) {
+    if (oggz_table_remove (ovdata->missing_eos, serialno) == -1) {
+      ret = log_error ();
+      fprintf (stderr, "serialno %010ld: *** eos marked but no bos\n",
+	       serialno);
+    }
+  }
+
   timestamp = gp_to_time (oggz, serialno, op->granulepos);
   if (timestamp != -1.0) {
     if (timestamp < current_timestamp) {
@@ -155,7 +244,7 @@ read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
     flush = OGGZ_FLUSH_AFTER;
   }
 
-  if ((feed_err = oggz_write_feed (writer, op, serialno, flush, NULL)) != 0) {
+  if ((feed_err = oggz_write_feed (ovdata->writer, op, serialno, flush, NULL)) != 0) {
     ret = log_error ();
     if (timestamp == -1.0) {
       fprintf (stderr, "%ld", oggz_tell (oggz));
@@ -176,46 +265,59 @@ read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
     }
   }
 
+  /* If this is the last packet in a chain, reset ovdata */
+  if (op->e_o_s && oggz_table_size (ovdata->missing_eos) == 0) {
+    ovdata_clear (ovdata);
+    ovdata_init (ovdata);
+  }
+
   return ret;
 }
 
 static int
 validate (char * filename)
 {
-  OGGZ * reader, * writer;
+  OGGZ * reader;
+  OVData ovdata;
   unsigned char buf[1024];
-  long n;
+  long n, nout = 0, bytes_written = 0;
   int active = 1;
 
   current_filename = filename;
-  current_timestamp = 0.0;
+  current_timestamp = 0;
   nr_errors = 0;
 
   /*printf ("oggz-validate: %s\n", filename);*/
-  
+
   if ((reader = oggz_open (filename, OGGZ_READ|OGGZ_AUTO)) == NULL) {
     fprintf (stderr, "oggz-validate: unable to open file %s\n", filename);
     exit (1);
   }
 
-  if ((writer = oggz_new (OGGZ_WRITE|OGGZ_AUTO)) == NULL) {
-    fprintf (stderr, "oggz-validate: unable to create new writer\n");
-    exit (1);
-  }
+  ovdata_init (&ovdata);
 
-  oggz_set_read_callback (reader, -1, read_packet, writer);
+  oggz_set_read_callback (reader, -1, read_packet, &ovdata);
+  oggz_set_read_page (reader, -1, read_page, &ovdata);
 
   while (active && (n = oggz_read (reader, 1024)) != 0) {
     if (nr_errors > MAX_ERRORS) {
       fprintf (stderr,
 	       "oggz-validate: maximum error count reached, bailing out ...\n");
       active = 0;
-    } else while (oggz_write_output (writer, buf, n) > 0);
+    } else while ((nout = oggz_write_output (ovdata.writer, buf, n)) > 0) {
+      bytes_written += nout;
+    }
   }
 
-  oggz_close (writer);
   oggz_close (reader);
 
+  if (bytes_written == 0) {
+    log_error ();
+    fprintf (stderr, "File contains no Ogg packets\n");
+  }
+
+  ovdata_clear (&ovdata);
+
   return active ? 0 : -1;
 }
 
