@@ -47,24 +47,29 @@
 #include <errno.h>
 
 #include <oggz/oggz.h>
+#include "oggz_tools.h"
 
-#define DEBUG
+/*#define DEBUG*/
 
 #undef MIN
 #define MIN(a,b) (((a)<(b))?(a):(b))
 
+typedef struct {
+  OggzTable * serialno_table;
+  OggzTable * content_types_table;
+  OggzReadPacket read_packet;
+  int dump_bits;
+  int dump_char;
+  int hide_offset;
+  int hide_serialno;
+  int hide_granulepos;
+  int hide_packetno;
+} ODData;
+
 static char * progname;
 static FILE * outfile = NULL;
-static int dump_bits = 0;
-static int dump_char = 1;
-static int dump_all_serialnos = 1;
 static int truth = 1;
 
-static int hide_offset = 0;
-static int hide_serialno = 0;
-static int hide_granulepos = 0;
-static int hide_packetno = 0;
-
 static void
 usage (char * progname)
 {
@@ -75,6 +80,10 @@ usage (char * progname)
   printf ("  -x, --hexadecimal      Generate a hexadecimal dump of each packet\n");
   printf ("\nFiltering options\n");
   printf ("  -n, --new              Only dump the first packet of each logical bitstream\n");
+  printf ("  -c content-type, --content-type content-type\n");
+  printf ("                         Dump only the logical bitstreams for a specified\n");
+  printf ("                         content-type. The following codecs are currently\n");
+  printf ("                         detected: theora, vorbis, speex, cmml\n");
   printf ("  -s serialno, --serialno serialno\n");
   printf ("                         Dump only the logical bitstream with specified serialno\n");
   printf ("  -O, --hide-offset      Hide the byte offset of each packet\n");
@@ -90,7 +99,36 @@ usage (char * progname)
   printf ("  -h, --help             Display this help and exit\n");
   printf ("  -v, --version          Output version information and exit\n");
   printf ("\n");
-  printf ("Please report bugs to <libannodex-dev@lists.annodex.net>\n");
+  printf ("Please report bugs to <ogg-dev@xiph.org>\n");
+}
+
+static ODData *
+oddata_new ()
+{
+  ODData * oddata = malloc (sizeof (ODData));
+  memset (oddata, 0, sizeof (ODData));
+
+  oddata->serialno_table = oggz_table_new ();
+  oddata->content_types_table = oggz_table_new ();
+
+  oddata->dump_bits = 0;
+  oddata->dump_char = 1;
+
+  oddata->hide_offset = 0;
+  oddata->hide_serialno = 0;
+  oddata->hide_granulepos = 0;
+  oddata->hide_packetno = 0;
+
+  return oddata;
+}
+
+static void
+oddata_delete (ODData * oddata)
+{
+  oggz_table_delete (oddata->serialno_table);
+  oggz_table_delete (oddata->content_types_table);
+
+  free (oddata);
 }
 
 static void
@@ -108,7 +146,7 @@ dump_char_line (unsigned char * buf, long n)
 }
 
 static void
-hex_dump (unsigned char * buf, long n)
+hex_dump (unsigned char * buf, long n, int dump_char)
 {
   int i;
   long remaining = n, count = 0;
@@ -146,7 +184,7 @@ hex_dump (unsigned char * buf, long n)
 }
 
 static void
-bin_dump (unsigned char * buf, long n)
+bin_dump (unsigned char * buf, long n, int dump_char)
 {
   int i, j;
   long remaining = n, count = 0;
@@ -199,12 +237,44 @@ bin_dump (unsigned char * buf, long n)
 static int
 read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
 {
-  fprintf (outfile, "%08" PRI_off_t "x: serialno %010ld, "
-	   "granulepos %" PRId64 ", packetno %" PRId64,
-	   hide_offset ? -1 : oggz_tell (oggz),
-	   hide_serialno ? -1 : serialno,
-	   hide_granulepos ? -1 : op->granulepos,
-	   hide_packetno ? -1 : op->packetno);
+  ODData * oddata = (ODData *) user_data;
+  ogg_int64_t units;
+  double time_offset;
+  int granuleshift;
+
+  if (oddata->hide_offset) {
+    fprintf (outfile, "oOo");
+  } else {
+    units = oggz_tell_units (oggz);
+    if (units == -1) {
+      fprintf (outfile, "%08" PRI_off_t "x", oggz_tell (oggz));
+    } else {
+      time_offset = (double)units / 1000.0;
+      ot_fprint_time (outfile, time_offset);
+    }
+  }
+
+  fprintf (outfile, ": serialno %010ld, ",
+	   oddata->hide_serialno ? -1 : serialno);
+
+  if (oddata->hide_granulepos) {
+    fprintf (outfile, "granulepos gGg,");
+  } else {
+    granuleshift = oggz_get_granuleshift (oggz, serialno);
+    if (granuleshift < 1) {
+      fprintf (outfile, "granulepos %" PRId64 ",", op->granulepos);
+    } else {
+      ogg_int64_t iframe, pframe;
+      iframe = op->granulepos >> granuleshift;
+      pframe = op->granulepos - (iframe << granuleshift);
+
+      fprintf (outfile, "granulepos %" PRId64 "|%" PRId64 ",",
+	       iframe, pframe);
+    }
+  }
+
+  fprintf (outfile, " packetno %" PRId64,
+	   oddata->hide_packetno ? -1 : op->packetno);
 
   if (op->b_o_s) {
     fprintf (outfile, " *** bos");
@@ -214,12 +284,14 @@ read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
     fprintf (outfile, " *** eos");
   }
 
-  fprintf (outfile, ":\n");
+  fprintf (outfile, ": ");
+  ot_fprint_bytes (outfile, op->bytes);
+  fputc ('\n', outfile);
 
-  if (dump_bits) {
-    bin_dump (op->packet, op->bytes);
+  if (oddata->dump_bits) {
+    bin_dump (op->packet, op->bytes, oddata->dump_char);
   } else {
-    hex_dump (op->packet, op->bytes);
+    hex_dump (op->packet, op->bytes, oddata->dump_char);
   }
 
   fprintf (outfile, "\n");
@@ -227,6 +299,27 @@ read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
   return 0;
 }
 
+static int
+filter_page (OGGZ * oggz, const ogg_page * og, long serialno, void * user_data)
+{
+  ODData * oddata = (ODData *) user_data;
+  const char * ident;
+  int i, n;
+
+  if (ogg_page_bos ((ogg_page *)og)) {
+    ident = ot_page_identify (og, NULL);
+    n = oggz_table_size (oddata->content_types_table);
+    for (i = 0; i < n; i++) {
+      char * c = oggz_table_nth (oddata->content_types_table, i, NULL);
+      if (strcasecmp (c, ident) == 0) {
+	oggz_set_read_callback (oggz, serialno, oddata->read_packet, oddata);
+      }
+    }
+  }
+
+  return 0;
+}
+
 static int
 ignore_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
 {
@@ -248,9 +341,11 @@ revert_file (char * infilename)
   OGGZ * oggz;
   FILE * infile;
   char line[80];
+  int hh, mm, ss;
   unsigned int offset;
   long current_serialno = -1, serialno;
-  ogg_int64_t granulepos, packetno;
+  ogg_int64_t granulepos, iframe, pframe, packetno;
+  int is_packetinfo = 0;
   int bos = 0, eos = 0;
 
   int line_offset = 0, consumed = 0;
@@ -270,41 +365,61 @@ revert_file (char * infilename)
     infile = fopen (infilename, "rb");
   }
 
-  oggz = oggz_new (OGGZ_WRITE|OGGZ_NONSTRICT);
+  oggz = oggz_new (OGGZ_WRITE|OGGZ_NONSTRICT|OGGZ_AUTO);
 
   while (fgets (line, 80, infile)) {
-    if (sscanf (line, "%x: serialno %ld, granulepos %lld, packetno %lld%n",
+    line_offset = 0;
+
+    /* Skip time offsets, OR ensure line_offset is 0 */
+    if (sscanf (line, "%2d:%2d:%2d.%n", &hh, &mm, &ss, &line_offset) < 3)
+      line_offset = 0;
+
+    is_packetinfo = 0;
+    if (sscanf (&line[line_offset], "%x: serialno %ld, granulepos %lld, packetno %lld%n",
 		&offset, &serialno, &granulepos, &packetno,
 		&line_offset) >= 4) {
+      is_packetinfo = 1;
+    } else {
+      if (sscanf (&line[line_offset], "%x: serialno %ld, granulepos %lld|%lld, packetno %lld%n",
+		  &offset, &serialno, &iframe, &pframe, &packetno,
+		  &line_offset) >= 5) {
+	int granuleshift = oggz_get_granuleshift (oggz, serialno);
+	is_packetinfo = 1;
+	granulepos = (iframe<<granuleshift)+pframe;
+      }
+    }
 
+    if (is_packetinfo) {
       /* flush any existing packets */
       if (current_serialno != -1) {
 	int ret;
 
 #ifdef DEBUG
-	printf ("feeding packet (%010ld) %ld bytes\n",
-		current_serialno, op.bytes);
+	printf ("feeding packet (%010ld) %ld bytes %s\n",
+		current_serialno, op.bytes,
+		op.b_o_s ? "bos" : "not bos");
 #endif
 	if ((ret = oggz_write_feed (oggz, &op, current_serialno, flush, NULL)) != 0) {
 	  fprintf (stderr, "%s: oggz_write_feed error %d\n", progname, ret);
 	}
 
 	while ((n = oggz_write_output (oggz, buf, 1024)) > 0) {
-	  fwrite (buf, 1, 1024, outfile);
+	  fwrite (buf, 1, n, outfile);
 	}
       }
 
       /* Start new packet */
       bos = 0; eos = 0;
-      if (sscanf (&line[line_offset], " *** %[b]%[o]%[s]%n", &c, &c, &c,
-		  &consumed) >= 3) {
-	bos = 1;
-	line_offset += consumed;
-      }
-      if (sscanf (&line[line_offset], " *** %[e]%[o]%[s]%n", &c, &c, &c,
-		  &consumed) >= 3) {
-	eos = 1;
-	line_offset += consumed;
+      while ((c=line[line_offset++]) && c != '*');
+      if (c == '*') {
+	if (!strncmp (&line[line_offset], "** bos", 6)) {
+	  bos = 1;
+	  line_offset += 6;
+	}
+	if (!strncmp (&line[line_offset], "** eos", 6)) {
+	  eos = 1;
+	  line_offset += 6;
+	}
       }
 
       current_serialno = serialno;
@@ -368,15 +483,19 @@ main (int argc, char ** argv)
   int show_version = 0;
   int show_help = 0;
 
+  ODData * oddata;
   OGGZ * oggz;
   char * infilename = NULL, * outfilename = NULL;
   int revert = 0;
-  OggzTable * table = NULL;
   long serialno;
-  OggzReadPacket my_read_packet = read_packet;
   int i, size;
   long n;
 
+  int filter_serialnos = 0;
+  int filter_content_types = 0;
+
+  ot_init ();
+
   progname = argv[0];
 
   if (argc < 2) {
@@ -384,10 +503,12 @@ main (int argc, char ** argv)
     return (1);
   }
 
-  table = oggz_table_new();
+  oddata = oddata_new ();
+
+  oddata->read_packet = read_packet;
 
   while (1) {
-    char * optstring = "hvbxnro:s:OSGP";
+    char * optstring = "hvbxnro:s:c:OSGP";
 
 #ifdef HAVE_GETOPT_LONG
     static struct option long_options[] = {
@@ -399,6 +520,7 @@ main (int argc, char ** argv)
       {"revert", no_argument, 0, 'r'},
       {"output", required_argument, 0, 'o'},
       {"serialno", required_argument, 0, 's'},
+      {"content-type", required_argument, 0, 'c'},
       {"hide-offset", no_argument, 0, 'O'},
       {"hide-serialno", no_argument, 0, 'S'},
       {"hide-granulepos", no_argument, 0, 'G'},
@@ -424,10 +546,10 @@ main (int argc, char ** argv)
       show_version = 1;
       break;
     case 'b': /* binary */
-      dump_bits = 1;
+      oddata->dump_bits = 1;
       break;
     case 'n': /* new */
-      my_read_packet = read_new_packet;
+      oddata->read_packet = read_new_packet;
       break;
     case 'o': /* output */
       outfilename = optarg;
@@ -436,21 +558,26 @@ main (int argc, char ** argv)
       revert = 1;
       break;
     case 's': /* serialno */
-      dump_all_serialnos = 0;
+      filter_serialnos = 1;
       serialno = atol (optarg);
-      oggz_table_insert (table, serialno, &truth);
+      oggz_table_insert (oddata->serialno_table, serialno, &truth);
+      break;
+    case 'c': /* content-type */
+      filter_content_types = 1;
+      n = (long)oggz_table_size (oddata->content_types_table);
+      oggz_table_insert (oddata->content_types_table, (long)n, optarg);
       break;
     case 'O': /* hide offset */
-      hide_offset = 1;
+      oddata->hide_offset = 1;
       break;
     case 'S': /* hide serialno */
-      hide_serialno = 1;
+      oddata->hide_serialno = 1;
       break;
     case 'G': /* hide granulepos */
-      hide_granulepos = 1;
+      oddata->hide_granulepos = 1;
       break;
     case 'P': /* hide packetno */
-      hide_packetno = 1;
+      oddata->hide_packetno = 1;
       break;
     default:
       break;
@@ -488,7 +615,7 @@ main (int argc, char ** argv)
   }
 
   if (revert) {
-    if (dump_bits) {
+    if (oddata->dump_bits) {
       fprintf (stderr, "%s: Revert of binary dump not supported\n", progname);
       goto exit_err;
     }
@@ -498,9 +625,9 @@ main (int argc, char ** argv)
     errno = 0;
 
     if (strcmp (infilename, "-") == 0) {
-      oggz = oggz_open_stdio (stdin, OGGZ_READ);
+      oggz = oggz_open_stdio (stdin, OGGZ_READ|OGGZ_AUTO);
     } else {
-      oggz = oggz_open (infilename, OGGZ_READ);
+      oggz = oggz_open (infilename, OGGZ_READ|OGGZ_AUTO);
     }
 
     if (oggz == NULL) {
@@ -514,13 +641,19 @@ main (int argc, char ** argv)
       goto exit_err;
     }
 
-    if (dump_all_serialnos) {
-      oggz_set_read_callback (oggz, -1, my_read_packet, NULL);
+    if (!filter_serialnos && !filter_content_types) {
+      oggz_set_read_callback (oggz, -1, oddata->read_packet, oddata);
     } else {
-      size = oggz_table_size (table);
-      for (i = 0; i < size; i++) {
-	oggz_table_nth (table, i, &serialno);
-	oggz_set_read_callback (oggz, serialno, my_read_packet, NULL);
+      if (filter_serialnos) {
+	size = oggz_table_size (oddata->serialno_table);
+	for (i = 0; i < size; i++) {
+	  oggz_table_nth (oddata->serialno_table, i, &serialno);
+	  oggz_set_read_callback (oggz, serialno, oddata->read_packet, oddata);
+	}
+      }
+      
+      if (filter_content_types) {
+	oggz_set_read_page (oggz, -1, filter_page, oddata);
       }
     }
 
@@ -530,10 +663,10 @@ main (int argc, char ** argv)
   }
 
  exit_ok:
-  oggz_table_delete (table);
+  oddata_delete (oddata);
   exit (0);
 
  exit_err:
-  oggz_table_delete (table);
+  oddata_delete (oddata);
   exit (1);
 }
