@@ -47,6 +47,8 @@
 #  define PRId64 "I64d"
 #endif
 
+#define READ_BLOCKSIZE 1024000
+
 static void
 usage (char * progname)
 {
@@ -164,7 +166,8 @@ oggzinfo_trackinfo_new (void)
 static long
 oi_bitrate (long bytes, ogg_int64_t ms)
 {
-  return (long) (((ogg_int64_t)bytes * 8 * 1000) / ms);
+  if (ms == 0) return 0;
+  else return (long) (((ogg_int64_t)bytes * 8 * 1000) / ms);
 }
 
 static void
@@ -349,7 +352,7 @@ oi_pass1 (OGGZ * oggz, OI_Info * info)
   oggz_set_read_page (oggz, -1, read_page_pass1, info);
   oggz_set_read_callback (oggz, -1, read_packet_pass1, info);
 
-  while ((n = oggz_read (oggz, 1024)) > 0);
+  while ((n = oggz_read (oggz, READ_BLOCKSIZE)) > 0);
 
   oggzinfo_apply (oit_calc_average, info);
 
@@ -365,7 +368,7 @@ oi_pass2 (OGGZ * oggz, OI_Info * info)
   oggz_set_read_page (oggz, -1, read_page_pass2, info);
   oggz_set_read_callback (oggz, -1, read_packet_pass2, info);
 
-  while ((n = oggz_read (oggz, 1024)) > 0);
+  while ((n = oggz_read (oggz, READ_BLOCKSIZE)) > 0);
 
   oggzinfo_apply (oit_calc_stddev, info);
 
@@ -493,7 +496,6 @@ main (int argc, char ** argv)
     
     oi_pass1 (oggz, &info);
 
-    oggz_seek_units (oggz, 0, SEEK_END);
     info.duration = oggz_tell_units (oggz);
     
     oi_pass2 (oggz, &info);
